/* ----------------------------------------------------------------------- */
/*  SequenceLib.c : implementation of the sequence library                 */
/*  Author : Ji HongKai ; Time: 2004.07                                    */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"
#include "time.h"

#include "MathLib.h"
#include "MatrixLib.h"
#include "StringLib.h"
#include "RandomLib.h"
#include "SequenceLib.h"

/* ----------------------------------------------------------------------------- */ 
/*                       struct tagSequence *SequenceCreate()                    */
/* This function is used for creating a sequence structure.                      */
/* ----------------------------------------------------------------------------- */ 
struct tagSequence *SequenceCreate()
{
	struct tagSequence *pNewSequence;

	/* Init */
	pNewSequence = NULL;

	/* new */
	pNewSequence = (struct tagSequence *)malloc(sizeof(struct tagSequence));
	if( pNewSequence == NULL )
		return NULL;
	pNewSequence->m_nIndex = -1;
	pNewSequence->m_nLength = 0;
	pNewSequence->m_pSequence = NULL;
	strcpy(pNewSequence->m_strAlias, "");
	pNewSequence->m_nStart = 0;
	pNewSequence->m_nEnd = 0;
	pNewSequence->m_pNext = NULL;

	/* return */
	return pNewSequence;
}

/* ----------------------------------------------------------------------------- */ 
/*                           void SequenceDelete()                               */
/* This function is used for deleting a sequence.                                */
/* ----------------------------------------------------------------------------- */ 
void SequenceDelete(struct tagSequence *pDelSequence)
{
	if( pDelSequence != NULL )
	{
		pDelSequence->m_pNext = NULL;
		DeleteString(pDelSequence->m_pSequence);
		free(pDelSequence);
	}	
}

/* ----------------------------------------------------------------------------- */ 
/*                            int  SequenceAddTail()                             */
/* This function is used for adding a segment to the tail of a sequence.         */
/* If the process is successful, it will return 1;                               */
/* else it will return 0.                                                        */            
/* ----------------------------------------------------------------------------- */ 
int SequenceAddTail(struct tagSequence *pSequence, char strLine[])
{
	/* nLen is used for recording the length of a string */
	int nLen;
	/* pString is used for manipulating the string */
	char *pString,*pTempString;

	/* check parameter */
	if(pSequence == NULL)
		return PROC_FAILURE;
	nLen = (int)strlen(strLine);
	if(nLen == 0)
		return PROC_SUCCESS;

	/* add */
	if(pSequence->m_pSequence == NULL)
	{
		pSequence->m_pSequence = CreateString(nLen);
		if(pSequence->m_pSequence == NULL)
		{
			printf("Error in SequenceAddTail: cannot create string file...\n");
			return PROC_FAILURE;
		}
		pString = pSequence->m_pSequence->m_pString;
		strcat(pString, strLine);
		pSequence->m_nLength = nLen;
		pSequence->m_nStart = 1;
		pSequence->m_nEnd = nLen;
	}
	else
	{
		pSequence->m_pSequence->m_nLength += nLen;
		pString = pSequence->m_pSequence->m_pString;
		pTempString = (char *)malloc(sizeof(char)*(pSequence->m_pSequence->m_nLength+1));
		if(pTempString == NULL)
			return PROC_FAILURE;
		strcpy(pTempString, pString);
		free(pString);
		strcat(pTempString, strLine);
		pSequence->m_pSequence->m_pString = pTempString;
		pSequence->m_nLength += nLen;
		pSequence->m_nEnd += nLen;
	}

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------------- */ 
/*                           void SequenceListClear()                            */
/* This function is used for clearing sequence list.                             */
/* ----------------------------------------------------------------------------- */ 
void SequenceListClear(struct tagSequence **pSequenceList)
{
	struct tagSequence *pHeadNode;

	/* clear sequence list. */
	pHeadNode = *pSequenceList;
	while( pHeadNode != NULL )
	{
		*pSequenceList = pHeadNode->m_pNext;
		SequenceDelete(pHeadNode);
		pHeadNode = *pSequenceList;
	}
}


/* ----------------------------------------------------------------------------- */ 
/*                          int SequenceListGetSize()                            */
/* This function is used for getting the size of a sequence list.                */
/* ----------------------------------------------------------------------------- */ 
int SequenceListGetSize(struct tagSequence *pSequenceList)
{
	int nCount;
	struct tagSequence *pSeq;

	/* init */
	nCount = 0;

	/* counting */
	pSeq = pSequenceList;
	while( pSeq != NULL )
	{
		nCount++;
		pSeq = pSeq->m_pNext;
	}

	/* return */
	return nCount;
}

/* ----------------------------------------------------------------------------- */ 
/*                     struct tagSequence *SequenceListGetAt()                   */
/* This function is used for getting a sequence from a sequence list.            */
/* ----------------------------------------------------------------------------- */ 
struct tagSequence *SequenceListGetAt(struct tagSequence *pSequenceList, int nIndex)
{
	int nCount;
	struct tagSequence *pSeq,*pTarget;

	/* init */
	nCount = SequenceListGetSize(pSequenceList);
	if((nIndex < 0) || (nIndex >= nCount))
	{
		printf("Error in struct tagSequence *SequenceListGetAt(): array index out of bound...\n");
		exit(EXIT_FAILURE);
	}

	pTarget = NULL;
	nCount = 0;

	/* searching */
	pSeq = pSequenceList;
	while( pSeq != NULL )
	{
		if(nCount == nIndex)
		{
			pTarget = pSeq;
			break;
		}
		nCount++;
		pSeq = pSeq->m_pNext;
	}

	/* return */
	return pTarget;
}

/* ----------------------------------------------------------------------------- */ 
/*                   struct tagSequence *SequenceListGetTail()                   */
/* This function is used for getting the tail sequence from a sequence list.     */
/* ----------------------------------------------------------------------------- */ 
struct tagSequence *SequenceListGetTail(struct tagSequence *pSequenceList)
{
	struct tagSequence *pSeq,*pTarget;

	/* searching */
	pTarget = NULL;
	pSeq = pSequenceList;
	while( pSeq != NULL )
	{
		pTarget = pSeq;		
		pSeq = pSeq->m_pNext;
	}

	/* return */
	return pTarget;
}


/* ----------------------------------------------------------------------------- */ 
/*                      int LoadFullSequenceList()                               */
/* This function loads all sequences from the input FASTA file.                  */
/* If the process is successful, it will return the count of sequences;          */
/* else it will return 0;                                                        */            
/* ----------------------------------------------------------------------------- */
int LoadFullSequenceList(char strInFilePath[], struct tagSequence **pSeqList)
{
	/* define */
	FILE *fpInput;
	char strLine[LONG_LINE_LENGTH];
	struct tagSequence *pNewSeq;
	struct tagSequence *pTailSeq;
	int nCount;
	
	/* get tail seq */
	pNewSeq = NULL;
	nCount = 0;
	if(*pSeqList == NULL)
		pTailSeq = NULL;
	else
	{
		pTailSeq = SequenceListGetTail(*pSeqList);
	}

	/* load file */
	fpInput = NULL;
	fpInput = fopen(strInFilePath, "rt");
	if(fpInput == NULL)
	{
        printf("Error in LoadConserveSequenceList: Cannot open input file...\n");
		return 0;
	}
	
	/* read */
	while(fgets(strLine, LONG_LINE_LENGTH, fpInput) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		if(strLine[0] == '>')
		{
			if(pNewSeq != NULL)
			{
				if(pTailSeq == NULL)
				{
					*pSeqList = pNewSeq;
					pTailSeq = pNewSeq;
				}
				else
				{
					pTailSeq->m_pNext = pNewSeq;
					pTailSeq = pNewSeq;
				}

				nCount++;
			}

			pNewSeq = NULL;
			pNewSeq = SequenceCreate();
			if(pNewSeq == NULL)
			{
				fclose(fpInput);
				printf("Error in LoadConserveSequenceList: Cannot create a new sequence node...\n");
				return 0;
			}

			pNewSeq->m_nIndex = nCount;
			strcpy(pNewSeq->m_strAlias, (strLine+1));
		}
		else
		{
			if(SequenceAddTail(pNewSeq, strLine) == PROC_FAILURE)
			{
				fclose(fpInput);
				printf("Error in LoadConserveSequenceList: cannot add string to sequence...\n");
				return 0;
			}			
		}
	}

	if(pNewSeq != NULL)
	{
		if(pTailSeq == NULL)
		{
			*pSeqList = pNewSeq;
			pTailSeq = pNewSeq;
		}
		else
		{
			pTailSeq->m_pNext = pNewSeq;
			pTailSeq = pNewSeq;
		}
		nCount++;
	}

	/* close files */
	fclose(fpInput);
	
	/* return */
	return nCount;
}

/* ----------------------------------------------------------------------------- */ 
/*                      FILE *SequenceWriteToFasta()                             */
/* This function is used for writing a sequences into a fasta file.              */
/* If the process is successful, it will return the end pointer of the file;     */
/* else it will return NULL;                                                     */
/* ----------------------------------------------------------------------------- */
FILE *SequenceWriteToFasta(struct tagSequence *pSequence, FILE *fpOut)
{
	/* nLinePos and pSeqPos are used for accessing the sequence. */
	int nLinePos;
	char *pSeqPos;

	/* check parameter */
	if(pSequence == NULL)
		return fpOut;
	if(fpOut == NULL)
		return NULL;

	/* write */
	if(pSequence->m_pSequence != NULL)
	{
		/* write head */
		fprintf(fpOut, ">%d\n", pSequence->m_nIndex);
		
		/* write sequence */
		nLinePos = 0;
		pSeqPos = pSequence->m_pSequence->m_pString;
		while(*pSeqPos != '\0')
		{
			fprintf(fpOut, "%c", *pSeqPos);
			nLinePos++;
			pSeqPos++;

			if(nLinePos == FASTA_LINE_LEN)
			{
				fprintf(fpOut, "\n");
				nLinePos = 0;
			}
		}
		if(nLinePos != 0)
		{
			fprintf(fpOut, "\n");
		}
	}

	/* return */
	return fpOut;
}

/* ----------------------------------------------------------------------------- */ 
/*                    SequenceWriteToFasta_ByStrand()                            */
/* This function is used for writing a sequences into a fasta file, according    */
/* to the strand specified.                                                      */
/* If the process is successful, it will return the end pointer of the file;     */
/* else it will return NULL;                                                     */
/* if nAliasType = 0: output the m_nIndex of the sequence as the header after >  */
/*                 1: output the strAlias as the header after >                  */
/* ----------------------------------------------------------------------------- */
FILE *SequenceWriteToFasta_ByStrand(struct tagSequence *pSequence, FILE *fpOut, 
									char chStrand, int nAliasType)
{
	/* nLinePos and pSeqPos are used for accessing the sequence. */
	int nLinePos;
	char *pSeqPos;
	int ni;
	char chBase;

	/* check parameter */
	if(pSequence == NULL)
		return fpOut;
	if(fpOut == NULL)
		return NULL;

	/* write */
	if(pSequence->m_pSequence != NULL)
	{
		/* write head */
		if(nAliasType == 1)
		{
			fprintf(fpOut, ">%s\n", pSequence->m_strAlias);
		}
		else
		{
			fprintf(fpOut, ">%d\n", pSequence->m_nIndex);
		}

		/* write sequence */
		if(chStrand == '-')
		{
			nLinePos = 0;
			pSeqPos = pSequence->m_pSequence->m_pString;
			for(ni=(pSequence->m_nLength-1); ni>=0; ni--)
			{
				switch(pSeqPos[ni])
				{
					case 'a': chBase = 't';
						break;
					case 'A': chBase = 'T';
						break;
					case 'c': chBase = 'g';
						break;
					case 'C': chBase = 'G';
						break;
					case 'g': chBase = 'c';
						break;
					case 'G': chBase = 'C';
						break;
					case 't': chBase = 'a';
						break;
					case 'T': chBase = 'A';
						break;
					case 'n': chBase = 'n';
						break;
					default: chBase = 'N';
				}

				fprintf(fpOut, "%c", chBase);
				nLinePos++;
				
				if(nLinePos == FASTA_LINE_LEN)
				{
					fprintf(fpOut, "\n");
					nLinePos = 0;
				}
			}
			if(nLinePos != 0)
			{
				fprintf(fpOut, "\n");
			}
		}
		else
		{
			nLinePos = 0;
			pSeqPos = pSequence->m_pSequence->m_pString;
			while(*pSeqPos != '\0')
			{
				fprintf(fpOut, "%c", *pSeqPos);
				nLinePos++;
				pSeqPos++;

				if(nLinePos == FASTA_LINE_LEN)
				{
					fprintf(fpOut, "\n");
					nLinePos = 0;
				}
			}
			if(nLinePos != 0)
			{
				fprintf(fpOut, "\n");
			}
		}
		
	}

	/* return */
	return fpOut;
}
